#include <string.h>

#include "ap_parser.h"

/* STATES:
 * 0 = in string
 * 1 = in script (inside)
 * 3 = script, beginning (so far, whitespace)
 * 4 = script, "%" found (candidate for ending)
 * 5 = string, "<" found

 */
enum parsing_state { 
    INSIDE_STRING	=	0,
    INSIDE_SCRIPT	=	1, 
    INSIDE_EXPRESSION	=	2, 
    BEGIN_SCRIPT	=	3,
    INSIDE_ENDING	=	4,
    INSIDE_BEGINING	=	5,
    INSIDE_EXP_ENDING	=	6
    };

#define MAX_LINES_TRANSLATED 1024
/* line_translation[<ln in target>]=<ln in src> */
static 
int
line_translation[MAX_LINES_TRANSLATED];

/* returns source line for given script line. 
Returns -1 if script line number is too big */
int co_translate_line(int num_script_line)
{
    if (num_script_line>=MAX_LINES_TRANSLATED) return -1;
    return line_translation[num_script_line];
}

/*
 * Parses text file containing embedded JavaScript 
 * like ( <% for (i=0; i<10; i++) {} %> )
 *
 * Requires allocated buffer strDest.
 * strDest should contain null-terminated string.
 * No more than sizeDest are copied.
 */
int co_ap_parse_template(char *strScript, const char *strSource, size_t sizeDest)
{
    size_t len;
    size_t sizeLeft;
    enum parsing_state numState = INSIDE_STRING
;
    int hasScripts = 0;
    int num_src_line=1;
    int num_dest_line=1;

    strncpy( strScript, "ap_write('", sizeDest );
    len = strlen( strScript );
    sizeLeft = sizeDest - len;
    strScript += len;

    memset(line_translation, 0, MAX_LINES_TRANSLATED);    

    /* in the loop we move at most one line at a time */
    while ((*strSource != '\0') && ( sizeLeft > 0 ))
    {
	if (*strSource == '\n') { num_src_line++; }
	/* not very effective: faster is to put it with num_dest_line increments */
	line_translation[num_dest_line]=num_src_line;
	switch (numState)
	{
	    case INSIDE_STRING :
		if (*strSource == '<')
		{
		    numState = INSIDE_BEGINING;
		}
		else
		{
		    /* copy current as part of js string literal */
		    switch (*strSource)
		    {
			case '\n'  : 
			    strncpy( strScript, "\\n", sizeLeft );
			    strScript += 2; sizeLeft -= 2;
            		    break;
			case '\r' : 
			    strncpy( strScript, "\\r", sizeLeft );
			    strScript += 2; sizeLeft -= 2;
            		    break;
			case '\t'  : 
			    strncpy( strScript, "\\t", sizeLeft );
			    strScript += 2; sizeLeft -= 2;
            		    break;
    			case '\''  : 
			case '"' : 
			case '\\' : 
			    *strScript++ = '\\'; sizeLeft--;
			    if (sizeLeft > 0)
			    {
				*strScript++ = *strSource; sizeLeft--;
			    }
			    break;
			default :
			    *strScript++ = *strSource; sizeLeft--;
		    }
		}
                break;
	    case INSIDE_BEGINING :
		if (*strSource == '%')
		{
		    numState = BEGIN_SCRIPT;
                    strncpy( strScript, "');\n", sizeLeft );
		    strScript += 3; sizeLeft -= 3;
		    num_dest_line++;
		}
		else
		{
		    if (*strSource != '<')
		    {
			numState = INSIDE_STRING;
			*strScript++ = '<'; sizeLeft--;
			if (sizeLeft > 0)
			{ /* copy but in string context!!! */
			    /* copy current as part of js string literal */
			    switch (*strSource)
			    {
				case '\n'  : 
				    strncpy( strScript, "\\n", sizeLeft );
				    strScript += 2; sizeLeft -= 2;
            			    break;
				case '\r' : 
				    strncpy( strScript, "\\r", sizeLeft );
				    strScript += 2; sizeLeft -= 2;
            			    break;
				case '\t'  : 
				    strncpy( strScript, "\\t", sizeLeft );
				    strScript += 2; sizeLeft -= 2;
            			    break;
    				case '\''  : 
				case '"' : 
				case '\\' : 
				    *strScript++ = '\\'; sizeLeft--;
				    if (sizeLeft > 0)
				    {
					*strScript++ = *strSource; sizeLeft--;
				    }
				    break;
				default :
				    *strScript++ = *strSource; sizeLeft--;
			    }
			}
		    }
		}
                break;
	    case BEGIN_SCRIPT :
		if (*strSource == '=')
		{
		    strncpy( strScript, "ap_write(", sizeLeft );
		    strScript += 9; sizeLeft -= 9;
		    numState = INSIDE_EXPRESSION;
		}
		else if (*strSource != ' ')
		{
		    numState = INSIDE_SCRIPT;
		    *strScript++ = *strSource; sizeLeft--;
		    if (*strSource == '\n') { num_dest_line++; }
		}
                break;
	    case INSIDE_SCRIPT :
		hasScripts = 1;
		if (*strSource == '%')
		{
		    numState = INSIDE_ENDING;
		}
		else
		{
		    *strScript++ = *strSource; sizeLeft--;
		    if (*strSource == '\n') { num_dest_line++; }
		}
                break;
	    case INSIDE_EXPRESSION :
		hasScripts = 1;
		if (*strSource == '%')
		{
		    numState = INSIDE_EXP_ENDING;
		}
		else
		{
		    *strScript++ = *strSource; sizeLeft--;
		    if (*strSource == '\n') { num_dest_line++; }
		}
                break;
	    case INSIDE_ENDING :
		if (*strSource == '>')
		{
		    numState = INSIDE_STRING;
                    strncpy( strScript, ";\nap_write('", sizeLeft );
		    strScript += 12; sizeLeft -= 12;
		    //num_dest_line++;
		}
		else
		{
    		    if (*strSource != '%')
		    {
			numState = INSIDE_SCRIPT;
			*strScript++ = '%'; sizeLeft--;
			if (sizeLeft > 0)
			{
			    *strScript++ = *strSource; sizeLeft--;
			    if (*strSource == '\n') { num_dest_line++; }
			}
		    }
		}
                break;
	    case INSIDE_EXP_ENDING :
		if (*strSource == '>')
		{
		    numState = INSIDE_STRING;
                    strncpy( strScript, ");\nap_write('", sizeLeft );
		    strScript += 13; sizeLeft -= 13;
		    //num_dest_line++;
		}
		else
		{
    		    if (*strSource != '%')
		    {
			numState = INSIDE_EXPRESSION;
			*strScript++ = '%'; sizeLeft--;
			if (sizeLeft > 0)
			{
			    *strScript++ = *strSource; sizeLeft--;
			    if (*strSource == '\n') { num_dest_line++; }
			}
		    }
		}
                break;
	}

	strSource++;
    }
    strncpy( strScript, "');", sizeLeft );
    return hasScripts;
}
